/* compilation under windows:
 *		cd <path>
 *		nasm -o calc.obj -f win --prefix _ calc.s
 *		gcc calc.c calc.obj -o calc.exe
 * compilation under linux (32 bit binary):
 *		cd <path>
 *		nasm -o calc.o -f elf32 calc.s
 *		gcc calc.c calc.o -o calc
 * for compiling a 32bit binary under 64bit linux the additional "-m32" for gcc switch is required
 */

#include <stdio.h>
#include <stdlib.h>

/*
 * The following two global variables are used to set epsilon_r1 and epsilon_r2 for the calculation.
 * They are declared in calc.s
 */
extern float epsilon_r1, epsilon_r2;

/**
 * This external function calculates the capacity of condensators to the given values
 *
 * @param int arrSize - how many sets of values have to be calculated
 * @param int* length - the pointer to an array of integers containing the length of the specific capacitor in mm
 * @param float* outer - the pointer to an array of floats containing the radius of the outer cope
 * @param float* inner - the pointer to an array of floats containing the radius of the inner cope
 * @param float* out1 - the pointer to an array of floats containing the results to the dielectricum 1
 * @param float* out2 - the pointer to an array of floats containing the results to the dielectricum 2
 */
extern void calc(int arrSize, int* lenghts, float* outer, float* inner, float* out1, float* out2);

///structure used as container for commandline arguments
struct Parameters {
	float er1, er2;
	FILE* in;
	FILE* out;
	int verbosity;
};

///prints a short help message
void usage(char *prgname);

///parses commandline arguments. results will be stored in the Parameters structure referenced by params.
void parse_params(struct Parameters* params, int argc, char**argv);

///a debug function used to print the input data to stderr
void debug_print_input(int size, int* in1, float* in2, float* in3);

///a debug function used to print the results to stderr
void debug_print_results(int size, float* out1, float* out2);

/**
 * Reads the input data from inFile.
 * This function tries to interpret the inFile as a tab-separated CSV file. If an error occurs, it will abort parsing
 * and print a warning to stderr. Everything read until this error, is stored correctly in the arrays.
 * This function is able to read an arbitrary amount of input data. It automaticly allocates memory. Therefore the pointers
 * referenced by in1, in2 and in3 will be overwritten. Make sure to free them after use. The number of data sets read will be
 * stored into size.
 */
void parse_csv(int* size, int** in1, float** in2, float** in3, FILE* inFile);

///writes the results as tab-separated CSV into the file referenced by out.
void write_csv(int size, float* out1, float* out2, FILE* out);

/**
 * main function.
 * return values:
 *	* 0: success
 *	* 1: parameter parsing failure
 *	* 2: memory allocation failure
 *	* 3: I/O error
 */
int main(int argc, char** argv)
{
	//create Parameters structure and set default values
	struct Parameters params;
	params.er1 = 2.2f; //paraffin
	params.er2 = 4.0f; //oilpaper
	params.in = stdin; //by default we read from stdin
	params.out = stdout; //and write to stdout
	params.verbosity = 0;
	//parse the command line
	parse_params(&params, argc, argv);

	//read input file
	int inputSize;
	int *in1;
	float *in2, *in3;
	if(params.verbosity >= 1) fputs(" * Parsing input file\n", stderr);
	parse_csv(&inputSize, &in1, &in2, &in3, params.in);
	//print the input data
	if(params.verbosity >= 2) debug_print_input(inputSize,in1,in2,in3);

	//allocate memory for results
	float* out1 = (float *)malloc(inputSize * sizeof(float));
	float* out2 = (float *)malloc(inputSize * sizeof(float));
	if(!out1 || !out2) {
		fputs("error: memory allocation failed\n", stderr);
		exit(2);
	}
	//set ers
	if(params.verbosity >= 1) fputs(" * Calculating...\n", stderr);
	if(params.verbosity >= 1) fprintf(stderr, "   -> er1: %f; er2: %f\n", params.er1, params.er2);
	epsilon_r1 = params.er1;
	epsilon_r2 = params.er2;
	//calculate the values
	calc(inputSize, in1, in3, in2, out1, out2);

	//print the results
	if(params.verbosity >= 2) debug_print_results(inputSize,out1,out2);
	if(params.verbosity >= 1) fputs(" * Writing results to file\n", stderr);
	//write to CSV
	write_csv(inputSize, out1, out2, params.out);

	if(params.verbosity >= 1) fputs(" * Cleaning up\n", stderr);
	//close the input file
	if(params.in != stdin) fclose(params.in);
	//and the output file
	if(params.out != stdout) fclose(params.out);
	//free the memory
	free(in1);
	free(in2);
	free(in3);
	free(out1);
	free(out2);

	return 0;
}

/*
 * For a description of the following functions see above.
 */

void usage(char *prgname)
{
	fprintf(stderr,"usage: %s [options] [input file] \n",prgname);
	fputs("The following options are avaible:\n", stderr);
	fputs(" -h\tprints this help\n", stderr);
	fputs(" -v <level>\tsets the verbosity to a level between 0 and 2; default:0\n", stderr);
	fputs(" -o <filename>\tsets the input file [default: stdin]\n", stderr);
	fputs("              \tWARNING: The current contents of this file will be overwritten\n", stderr);
	fputs(" -er1 <value>\tsets the value for epsilon_r 1\n", stderr);
	fputs(" -er2 <value>\tsets the value for epsilon_r 2\n", stderr);
	fputs("\n", stderr);
}

void parse_params(struct Parameters* params, int argc, char**argv) {
	//the argument currently handled
	int currArg = 1;
	//some booleans to make sure, that every parameter will be specified only once
	char outSpecified = 0, inSpecified = 0, vSpecified = 0, er1Specified = 0, er2Specified = 0;
	//handle all passed arguments
	while(currArg < argc) {
		//compare them with "-<char>" switches
		if(strcmp(argv[currArg], "-h") == 0) {
			//print usage information and exit
			usage(argv[0]);
			exit(0);
		} else if(strcmp(argv[currArg], "-o") == 0) {
			//=>output file => next argument: filename
			currArg++;
			//some sanity checks
			if(currArg >= argc) {
				fputs("error: filename expected after -o, but nothing found\n", stderr);
				usage(argv[0]);
				exit(1);
			}
			if(outSpecified) {
				fputs("error: output file already set\n", stderr);
				exit(1);
			}
			//open it
			params->out = fopen(argv[currArg],"w");
			if(!params->out) {
				fprintf(stderr, "error: unable to open output file: \"%s\"", argv[currArg]);
				usage(argv[0]);
				exit(3);
			}
			outSpecified = 1;
		} else if(strcmp(argv[currArg], "-v") == 0) {
			//verbosity => next argument: verbosity level
			currArg++;
			//check sanity
			if(currArg >= argc) {
				fputs("error: value expected after -v, but nothing found\n", stderr);
				usage(argv[0]);
				exit(1);
			}
			if(vSpecified) {
				fputs("error: verbosity already set\n", stderr);
				exit(1);
			}
			//parse the verbosity level (in a pedantic way)
			char* next_char;
			params->verbosity = strtod(argv[currArg], &next_char);
			if(*next_char != '\0') {
				fputs("error: invalid value for er1 specified", stderr);
				exit(1);
			}
			//range checking
			if(params->verbosity < 0 || params->verbosity > 2) {
				fputs("error: invalid verbosity level. allowed range: [0-2]\n", stderr);
				usage(argv[0]);
				exit(1);
			}
			vSpecified = 1;
		} else if(strcmp(argv[currArg], "-er1") == 0) {
			//er1 => next argument: value for epsilon_r1
			currArg++;
			//the usual checks... once again
			if(currArg >= argc) {
				fputs("error: value expected after -er1, but nothing found\n", stderr);
				usage(argv[0]);
				exit(1);
			}
			if(er1Specified) {
				fputs("error: er1 already set\n", stderr);
				exit(1);
			}
			//parse the value (and make sure that the whole string was parsed!)
			char* next_char;
			params->er1 = strtod(argv[currArg], &next_char);
			if(*next_char != '\0') {
				fputs("error: invalid value for er1 specified", stderr);
				exit(1);
			}
			er1Specified = 1;
		} else if(strcmp(argv[currArg], "-er2") == 0) {
			//er2 => see above
			currArg++;
			if(currArg >= argc) {
				fputs("error: value expected after -er2, but nothing found\n", stderr);
				usage(argv[0]);
				exit(1);
			}
			if(er2Specified) {
				fputs("error: er2 already set\n", stderr);
				exit(1);
			}
			char* next_char;
			params->er2 = strtod(argv[currArg], &next_char);
			if(*next_char != '\0') {
				fputs("error: invalid value for er2 specified", stderr);
				exit(1);
			}
			er2Specified = 1;
		} else {
			//not a switch? => input file
			if(inSpecified) {
				fputs("error: input file already set\n", stderr);
				exit(1);
			}
			//(try to) open it
			params->in = fopen(argv[currArg],"r");
			if(!params->in) {
				fprintf(stderr, "error: unable to open input file: \"%s\"", argv[currArg]);
				usage(argv[0]);
				exit(3);
			}
			inSpecified = 1;
		}
		//NEXT!
		currArg++;
	}
}

void debug_print_input(int size, int* in1, float* in2, float* in3) {
	fputs("input data:\n", stderr);
	int i; for(i = 0; i < size; i++) {
		fprintf(stderr, "[%i] =>\t%i\t%e\t%e\n", i, in1[i], (double) in2[i], (double) in3[i]);
	}
}

void debug_print_results(int size, float* out1, float* out2) {
	fputs("results:\n", stderr);
	int i; for(i = 0; i < size; i++) {
		fprintf(stderr, "[%i] =>\t%e\t%e\n", i, (double) out1[i], (double) out2[i]);
	}
}

void parse_csv(int* size, int** in1, float** in2, float** in3, FILE* inFile) {
	//reserve some initial memory
	*size = 50;
	*in1 = (int *)malloc(*size * sizeof(int));
	*in2 = (float *)malloc(*size * sizeof(float));
	*in3 = (float *)malloc(*size * sizeof(float));
	if(!*in1 || !*in2 || !*in3){
		fputs("error: memory allocation failed", stderr);
		exit(2);
	}
	//Parse the input file
 	int linesread = 0;
 	while(!feof(inFile)) {
		//buffer full?
		if(linesread>*size-1)
		{
			//double the size.
			//why not just add 50? => GAD. By doubling the size every time we get O(n) = n.
			//Otherwise we would get O(n) = n*n.
			*size *= 2;
			//the buffer contents will be copied by realloc automatically
			*in1 = (int *) realloc(*in1, *size * sizeof(int));
			*in2 = (float *) realloc(*in2, *size * sizeof(float));
			*in3 = (float *) realloc(*in3, *size * sizeof(float));
			if(!*in1 || !*in2 || !*in3) {
				fputs("error: memory allocation failed\n", stderr);
				exit(2);
			}
		}
		if(fscanf(inFile,"%d\t%f\t%f\n", 
		  &(*in1)[linesread], &(*in2)[linesread], &(*in3)[linesread]) != 3){
			fprintf(stderr, "Parsing input file: Error in line %i\n",linesread);
			fprintf(stderr, "Aborted.\n");
			break;
		}
		linesread++;
	}
	//tailor the buffers.
	*size = linesread;
	*in1 = (int *) realloc(*in1, *size * sizeof(int));
	*in2 = (float *) realloc(*in2, *size * sizeof(float));
	*in3 = (float *) realloc(*in3, *size * sizeof(float));
	if(!*in1 || !*in2 || !*in3) {
		fputs("error: memory allocation failed", stderr);
		exit(2);
	}
}

void write_csv(int size, float* out1, float* out2, FILE* out) {
	int i; for(i = 0; i < size; i++) {
		fprintf(out, "%e\t%e\n",out1[i], out2[i]);
	}
}

